#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "FileReader.h"

static int infd = -1;  // file descriptor (used by open(), read(), close())

// statistics counters.

static struct {
  int  close;
  int  open;
  int  read;
  int  seek;
  int  stat;
} statistic;

//-----------------------------------------------------------
// fileOpen()
//   Must be called before any fileRead() calls are made.
//   Argument is the name of a file.
//   Returns:  0 for success
//             non-zero for failure
//----------------------------------------------------------

int fileOpen( char* filename) {
  infd = open( filename, O_RDONLY );
  statistic.open++;
  if ( infd == -1 ) return errno;
  return 0;
}

//---------------------------------------------------
// fileClose()
//   Should be invoked after done with the store.
//   Gives it a chance to do any clean-up it might need.
//----------------------------------------------------

void fileClose() {
  if ( infd != -1 ) close(infd);
  statistic.close++;
  infd = -1;
}

//---------------------------------------------------
// fileRead()
//   Reads length bytes at address addr in the file,
//   putting them in buf.
//   Returns: 0 for success
//            non-zero for failure
//---------------------------------------------------

int fileRead( FileAddress addr, size_t length, void* buf ) {
  ssize_t nread;
  int rcode;

  rcode = lseek( infd, addr, SEEK_SET );
  statistic.seek++;
  if ( rcode == -1 ) return errno;  // some error has occurred

  nread = read( infd, buf, length );
  statistic.read++;

  if ( nread == -1 ) return errno;    // some error executing read
  if ( nread != length ) return -1;   // got only some of the data, which we consider an error

  return 0;
}

//----------------------------------------------------------
// fileSize()
//   Returns file size through the size argument.
//   Returns a success indicator:
//    0 - success
//    other - failed
//----------------------------------------------------------

int fileSize(size_t* size) {
  struct stat statBuf;
  int rcode;

  *size = 0;
  if ( infd == -1 ) return -1;  // file must be open
  rcode = fstat( infd, &statBuf );
  statistic.stat++;
  if ( rcode ) return errno;

  *size = statBuf.st_size;
  return 0;
}

//----------------------------------------------------------
// resetStatistics()
//   Sets all counters to 0
//----------------------------------------------------------

void fileResetStatistics() {
  statistic.close = 0;
  statistic.open = 0;
  statistic.read = 0;
  statistic.seek = 0;
  statistic.stat = 0;
}

//----------------------------------------------------------
// printStatistics()
//   Prints (to stdout) the (implementation specific) list
//   of counters.
//----------------------------------------------------------

void filePrintStatistics() {
  printf("Closes:\t%d\n", statistic.close);
  printf("Opens:\t%d\n", statistic.open);
  printf("Reads:\t%d\n", statistic.read);
  printf("Seeks:\t%d\n", statistic.seek);
  printf("Stats:\t%d\n", statistic.stat);
}
