#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <stdlib.h>
#include <string.h>
#include <regex.h>
#include <malloc.h>

#include "RecordStore.h"
#include "FileReader.h"
#include "printTree.h"
#include "artistCDs.h"
#include "listArtists.h"

//---------------------------------------------------
// Usage: ./a.out <filename>
//   e.g., ./a.out freedb-small.dat  640
//---------------------------------------------------

void usage(char* cmd) {
  fprintf(stderr, "Usage: %s <filename>\n", cmd );
  fprintf(stderr, "\tOnce running, 'artist foo' prints all CD information for artist foo.\n");
  fprintf(stderr, "\t'artists' lists all artist names.\n");
  fprintf(stderr, "\t'print' prints all file contents.\n");
}

//--------------------------------------------------------------------------
// main()
//  Sits in loop reading commands and calling a routine to execute each.
//--------------------------------------------------------------------------

int main( int argc, char* argv[] ) {
  RecordAddress rootAddress;
  int         rcode;

  char*       regstr;
  regex_t     re;
  regmatch_t* matches;

  char*       inputLine = NULL;
  char*       expansion = NULL;

  //-----------------------------
  // correctly invoked?
  //-----------------------------
  if ( argc != 2 ) {
    usage( argv[0] );
    return 1;  // error return
  }

  //----------------------------------------------------------------------------
  // initialize statistic keeping by the FileReader implementation
  //----------------------------------------------------------------------------
  fileResetStatistics();

  //----------------------------------------------------------------------------
  // try to initialize the record store.  On success, we get the root record address.
  //----------------------------------------------------------------------------
  rcode = storeOpen( argv[1], &rootAddress );
  if ( rcode ) {
    fprintf(stderr, "Error opening '%s': %s\n", argv[1], strerror(rcode));
    return rcode;
  }

  //----------------------------------------------------------------------------
  // prepare to parse command lines
  //   You can safely ignore this, but basically we set up a regular expression
  //   that matches lines that look like:
  //        print | artist any-string-including-embedded-whitespace | artists | empty-line
  //   We use this in the body of the loop to parse command lines type by the user.
  //----------------------------------------------------------------------------
  regstr = "(^\\s*(print)\\s*$)|(^\\s*(artist)\\s+([^\\s].*[^\\s]*)\\s*$)|(^\\s*(artists)\\s*$)|(^\\s*$)";
  rcode = regcomp( &re, regstr, REG_EXTENDED );
  if (rcode !=  0) {
    regerror( rcode, &re, regstr, sizeof(regstr) );
    fprintf( stderr, "Invalid regular expression: %s\n", regstr );
    regfree( &re );
    return 1;
  }

  matches = (regmatch_t *)malloc( (re.re_nsub + 1) * sizeof(regmatch_t) );
  if (!matches) {
    fputs( "Out of memory!\n", stderr );
    return 1;
  }

  using_history();

  //----------------------------------
  // now loop, handling commands
  //----------------------------------

  while ( 1 ) {

    // free up last iteration's buffers
    if ( inputLine ) { free(inputLine); inputLine = NULL; }
    if ( expansion ) { free(expansion); expansion = NULL; }

    // read an input line
    inputLine = readline("\ncommand> ");
    if ( !inputLine ) break;  // done on eof

    // perform history expansion
    rcode = history_expand(inputLine, &expansion);
    if ( rcode < 0 ) {
      fprintf(stderr, "History expansion error: '%s'\n", expansion);
      continue;
    }

    // echo line if expansion occurred
    if ( rcode==1 || rcode==2) {
      fprintf(stderr, "%s\n", expansion);
    }

    // if error or request to print history line, don't execute command
    if ( rcode < 0 || rcode == 2 ) continue;

    // add line to history
    if ( *expansion ) add_history(expansion);

    // now try to parse the line
    rcode = regexec( &re, expansion, re.re_nsub + 1, matches, 0 );

    if (rcode) {
      fprintf( stderr, "Unrecognized command: %s\n", expansion );
    }
    else  if ( matches[1].rm_so != -1 ) {
      // it's a print command
      rcode = printTree( rootAddress, 0 );
      if ( rcode )  {
	fprintf(stderr, "Error printing record tree rooted at %u: %s\n", rootAddress, strerror(rcode) );
      }
    }
    else if ( matches[4].rm_so != -1 ) {
      // it's an artist command
      char* artistName = expansion+matches[5].rm_so;
      RecordAddress printAddress = artistCDs(artistName, rootAddress );
      if ( !printAddress ) {
	fprintf(stderr, "Error finding artist '%s': %d\n", artistName, rcode );
	continue;
      }
      printf("Artist: '%s'\n\n", artistName);
      rcode = printTree( printAddress, 0 );
      if ( rcode )  {
	fprintf(stderr, "Error printing record tree rooted at %u: %s\n", printAddress, strerror(rcode) );
      }
    } else if ( matches[6].rm_so != -1 ) {
      // artists command
      listArtists( rootAddress );
    }

  } // while loop

  // clean up pattern matching storage
  free(matches);
  regfree( &re );

  // clean up command line processing storage
  clear_history();

  // all done...
  storeClose();

  // prints statistics kept by the FileReader implementation
  filePrintStatistics();

  return 0;
}

  
